#include "affine.h"
#include "math.h"

// Vectors:

void add_vectors(double *u,  double *v, double *out)
{ out[0] = u[0] + v[0];
  out[1] = u[1] + v[1];
  out[2] = u[2] + v[2];
}

void add_vectors_self(double *u, double *out)
{ out[0] += u[0];
  out[1] += u[1];
  out[2] += u[2];
}

void sub_vectors(double *u, double *v, double *out)
{ out[0] = u[0] - v[0];
  out[1] = u[1] - v[1];
  out[2] = u[2] - v[2];
}

void sub_vectors_self(double *u, double *out)
{ out[0] -= u[0];
  out[1] -= u[1];
  out[2] -= u[2];
}

void scale_vector(double s, double *u, double *out)
{ out[0] = s * u[0];
  out[1] = s * u[1];
  out[2] = s * u[2];
}

void scale_vector_self(double s, double *out)
{ out[0] *= s;
  out[1] *= s;
  out[2] *= s;
}

double dot_product(double *u, double *v)
{
  return u[0]*v[0] + u[1]*v[1] + u[2]*v[2];
}

void cross_product(double *u, double *v, double *out)
{
  out[0] = u[1]*v[2] - u[2]*v[1];
  out[1] = u[2]*v[0] - u[0]*v[2];
  out[2] = u[0]*v[1] - u[1]*v[0];
}

void cross_product_self(double *u, double *out)
{ double t0, t1;

  t0 = u[1]*out[2] - u[2]*out[1];
  t1 = u[2]*out[0] - u[0]*out[2];
  out[2] = u[0]*out[1] - u[1]*out[0];
  out[1] = t1;
  out[0] = t0;
}

double vector_length(double *u)
{ 
  return sqrt(u[0]*u[0] + u[1]*u[1] + u[2]*u[2]);
}

void unit_vector(double *u, double *out)
{
  double length = vector_length(u);
  if (length == 0)
  { out[0] = 0; out[1] = 0; out[2] = 0;
    return;
  }
  out[0] = u[0]/length;
  out[1] = u[1]/length;
  out[2] = u[2]/length;
}

void unit_vector_self(double *out)
{
  double length = vector_length(out);
  if (length == 0) return;
  out[0] /= length;
  out[1] /= length;
  out[2] /= length;
}

double distance_vectors(double *u, double *v)
{ 
  sub_vectors_self(u, v);  // v -= u;
  return vector_length(v); // length[u-v]
}

double angle_vectors(double *u, double *v)
{ double temp;

  temp = dot_product(u, v) / (vector_length(u)*vector_length(v));
  return acos(temp);
}

// projection of u on s
void project_vectors(double *u, double *s, double *out)
{
  double unit[3];
  unit[0] = s[0]; unit[1] = s[1]; unit[2] = s[1];
  unit_vector_self(unit); // length(unit) = 1
  scale_vector(dot_product(u, unit), unit, out);
}

// projection of out on s
void project_vectors(double *s, double *out)
{ double unit[3];
  unit[0] = s[0]; unit[1] = s[1]; unit[2] = s[1];
  unit_vector(s, unit); // length(unit) = 1
  scale_vector(dot_product(out, unit), unit, out);
}

// Matrix (Self versions are not particulary useful, because they use an
//         internal copy)

void mul_mat_vec(double *in, double *matrix, double *out)
{ 
  int i, j;

  for (i = 0; i < 3; i++)
  { 
    out[i] = 0;
	for (j = 0; j < 3; j++)
	  out[i] += matrix[i*3+j]*in[j];
  }
}

void mul_mat_vec_self(double *mat, double *out)
{ 
  int i, j;
  double temp[3];

  for (i = 0; i < 3; i++)
  { 
    temp[i] = 0;
	for (j = 0; j < 3; j++)
	  temp[i] += mat[i*3+j]*out[j];
  }
  out[0] = temp[0];
  out[1] = temp[1];
  out[2] = temp[2];
}

// Out = N * M
void mul_mat_mat(double *m, double *n, double *out)
{ 
  int i, j, k;

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
	{
	 out[i*3+j] = 0;
	 for (k = 0; k < 3; k++)
	   out[i*3+j] += n[i*3+k] * m[k*3+j];
	}
}

// out = N * OUT;
void mul_mat_mat_self(double *n, double *out)
{ 
  int i, j, k;
  double temp[9];

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
	{
	 temp[i*3+j] = 0;
	 for (k = 0; k < 3; k++)
	   temp[i*3+j] += n[i*3+k] * out[k*3+j];
	}

  for (i = 0; i < 9; i++) out[i] = temp[i];
}


// out = out*N;
void mul_mat_self_mat(double *n, double *out)
{ 
  int i, j, k;
  double temp[9];

  for (i = 0; i < 3; i++)
    for (j = 0; j < 3; j++)
	{
	 temp[i*3+j] = 0;
	 for (k = 0; k < 3; k++)
	   temp[i*3+j] += out[i*3+k] * n[k*3+j];
	}

  for (i = 0; i < 9; i++) out[i] = temp[i];
}

// Matrix Creation

void make_normal_mat(double* in)
{ 
  for (int i= 1; i < 8; i++) in[i] = 0;
  in[0] = 1;
  in[4] = 1;
  in[8] = 1;
}

void make_rotz_mat(double* in, double angle)
{
  make_normal_mat(in);
  in[0] = cos(angle);
  in[1] = -sin(angle);
  in[3] = sin(angle);
  in[4] = cos(angle);
}

void make_roty_mat(double* in, double angle)
{
  make_normal_mat(in);
  in[0] = cos(angle);
  in[2] = -sin(angle);
  in[6] = sin(angle);
  in[8] = cos(angle);
}

void make_rotx_mat(double* in, double angle)
{
  make_normal_mat(in);
  in[4] = cos(angle);
  in[5] = -sin(angle);
  in[7] = sin(angle);
  in[8] = cos(angle);
}

void make_scale_mat(double* in, double scale_x, double scale_y, double scale_z)
{ 
  for (int i= 0; i < 9; i++)
    in[i] = 0;

  in[0] = scale_x;
  in[4] = scale_y;
  in[8] = scale_z;
}

void make_scalex_mat(double* in, double scale)
{
	make_normal_mat(in);
	in[0] = scale;
}

void make_scaley_mat(double* in, double scale)
{
	make_normal_mat(in);
	in[4] = scale;
}

void make_scalez_mat(double* in, double scale)
{
	make_normal_mat(in);
	in[8] = scale;
}

void make_skewx_mat(double* in, double angle)
{
  make_normal_mat(in);
  in[4] = cos(angle/2);
  in[7] = -sin(angle/2);
  in[5] = -sin(angle/2);
  in[8] = cos(angle/2);
}

void make_skewy_mat(double* in, double angle)
{ 
  make_normal_mat(in);
  in[0] = cos(angle/2);
  in[6] = -sin(angle/2);
  in[2] = -sin(angle/2);
  in[8] = cos(angle/2);
}

void make_skewz_mat(double* in, double angle)
{
  make_normal_mat(in);
  in[0] = cos(angle/2);
  in[3] = -sin(angle/2);
  in[1] = -sin(angle/2);
  in[4] = cos(angle/2);
}

// Advanced stuff

// u1, u2 and u3 should be lineair independent
void Gram_Schmid_orthonormalize(double *v1, double *v2, double *v3)
{
  double  w2[3], w3[3], t1[3], t2[3];

  unit_vector_self(v1); // w1 = v1 = first unit vector
  
  w2[0] = v2[0]; w2[1] = v2[1]; w2[2] = v2[2];
  scale_vector(dot_product(v2, v1), v1, t1); // t1 = <v2,w1>w1
  sub_vectors_self(t1, w2);	// w2 = v2 - <v2,w1>w1
  unit_vector_self(w2);		// w2 = second unit vector

  w3[0] = v3[0]; w3[1] = v3[1]; w3[2] = v3[2];
  scale_vector(dot_product(v3, v1), v1, t1); // t1 = <v3,w1>w1
  scale_vector(dot_product(v3, w2), w2, t2); // t2 = <v3,w2>w2
  sub_vectors_self(t1, w3);	// w3 = v3 - <v3,w1>w1
  sub_vectors_self(t2, w3);	// w3 = v3 - (<v3,w1>w1 + <v3,w2>w2)
  unit_vector_self(w3);		// w3 = third unit vector

  v2[0] = w2[0]; v2[1] = w2[1]; v2[2] = w2[2];
  v3[0] = w3[0]; v3[1] = w3[1]; v3[2] = w3[2];
}


// u1, u2 and u3 should be lineair independent

// This SHOULD result in three orthogonal vectors,
// but without the scaling to unit vectors

void Gram_Schmid_orthogonize(double *v1, double *v2, double *v3)
{

  double  u1[3], u2[3], //u3[3], // Unit vectors
                 w2[3], w3[3], // result vectors
				 t1[3], t2[3]; // temp vectors

  unit_vector(v1, u1); // v1 = first result vector, u1 = first unit vector
  
  w2[0] = v2[0]; w2[1] = v2[1]; w2[2] = v2[2];
  scale_vector(dot_product(v2, u1), u1, t1); // t1 = <v2,u1>u1
  sub_vectors_self(t1, w2);	// w2 = v2 - <v2,w1>w1
  unit_vector(w2, u2);		// w2 = second result vector, u2 = second unit vector

  w3[0] = v3[0]; w3[1] = v3[1]; w3[2] = v3[2];
  scale_vector(dot_product(v3, u1), u1, t1); // t1 = <v3,u1>u1
  scale_vector(dot_product(v3, u2), u2, t2); // t2 = <v3,u2>u2

  sub_vectors_self(t1, w3);	// w3 = v3 - <v3,u1>u1
  sub_vectors_self(t2, w3);	// w3 = v3 - (<v3,u1>u1 + <v3,u2>u2)
//  unit_vector(w3, u3);		// w3 = third unit vector, u3 = third unit vector

  v2[0] = w2[0]; v2[1] = w2[1]; v2[2] = w2[2];
  v3[0] = w3[0]; v3[1] = w3[1]; v3[2] = w3[2];
}

