//
// IFS Fractals
//
//

const double EPS = 0.000001;

// include files
//#include <stdlib.h>

#include "screen.h"
#include "IFS_Interface.h"

double abs(double in)
{
	if (in >= 0)
		return in;
	else 
		return -in;
}

class QuadEdge;

class Point2d
{	
public:
	Point2d(float inx, float iny) { x = inx; y = iny; }
	bool operator==(const Point2d& in) const { return x== in.x && y == in.y; }
	float x, y;
};

class Edge
{
	friend QuadEdge;
	friend void Splice(Edge*, Edge*);
private:
	int	num;
	Edge* next;
	Point2d *data;
public:
	Edge() { data = 0; }

	Edge*	Rot();
	Edge*	invRot();
	Edge*	Sym();
	Edge*	Onext();
	Edge*	Oprev();
	Edge*	Dnext();
	Edge*	Dprev();
	Edge*	Lnext();
	Edge*	Lprev();
	Edge*	Rnext();
	Edge*	Rprev();
	Point2d* Org();
	Point2d* Dest();
	const Point2d& Org2d() const;
	const Point2d& Dest2d() const;
	void EndPoints(Point2d*, Point2d*);
	QuadEdge* Qedge() { return (QuadEdge *) (this - num); }
};

class QuadEdge {
	friend Edge *MakeEdge();
private:
	Edge e[4];
public:
	QuadEdge();
};

class Subdivision {
private:
	Edge* startingEdge;
	Edge* Locate(const Point2d&);
public:
	Subdivision(const Point2d&, const Point2d&, const Point2d&);
	void InsertSite(const Point2d&);
	void Draw();
};

inline QuadEdge::QuadEdge()
{
	e[0].num = 0; e[1].num = 1; e[2].num = 2; e[3].num = 3;
	e[0].next = &(e[0]); e[1].next = &(e[3]);
	e[2].next = &(e[2]); e[3].next = &(e[1]);
}

/**************************** Edge Algebra *****************************/

inline Edge* Edge::Rot()
// Return the dual of the current edge, directed from its right to its left
{
	return (num < 3)? this + 1: this - 3;
}

inline Edge* Edge::invRot()
// Return the dual of the current edge, directed from its left to its right
{
	return (num > 0)? this - 1: this + 3;
}

inline Edge* Edge::Sym()
// Return the edge from the destination to the origin of the current edge
{
	return (num < 2)? this + 2: this - 2;
}

inline Edge* Edge::Onext()
// Return next ccw edge around (from) the origin of the current edge
{
	return next;
}

inline Edge* Edge::Oprev()
// Return next cw edge around (from) the origin of the current edge
{
	return Rot()->Onext()->Rot();
}

inline Edge* Edge::Dnext()
// Return next ccw edge around (into) the destination of the current edge
{
	return Sym()->Onext()->Sym();
}

inline Edge* Edge::Dprev()
// Return next cw edge around (into) the destination of the current edge
{
	return invRot()->Onext()->invRot();
}

inline Edge* Edge::Lnext()
// Return the ccw edge around the left face following the current edge
{
	return invRot()->Onext()->Rot();
}

inline Edge* Edge::Lprev()
// Return the ccw edge around the left face before the current edge
{
	return Onext()->Sym();
}

inline Edge* Edge::Rnext()
// Return the edge around the right face ccw following the current edge
{
	return Rot()->Onext()->invRot();
}

inline Edge* Edge::Rprev()
// Return the edge around the right face ccw before the current edge
{
	return Sym()->Onext();
}

/**************************** Access to data pointers *****************************/

inline Point2d* Edge::Org()
{
	return data;
}

inline Point2d* Edge::Dest()
{
	return Sym()->data;
}

inline const Point2d& Edge::Org2d() const
{
	return *data;
}

inline const Point2d& Edge::Dest2d() const
{
	return (num < 2) ? *((this + 2)->data) : *((this - 2)->data);
}

inline void Edge::EndPoints(Point2d* or, Point2d* de)
{
	data = or;
	Sym()->data = de;
}

/**************************** Access to data pointers *****************************/

Edge* MakeEdge()
{
	QuadEdge *ql = new QuadEdge;
	return ql->e;
}

void Splice(Edge* a, Edge* b)
// This operator affects the two edge rings around the origins of a en b
// and, independently, the two edge rings around the left faces of a and b.
// In each case, (i) if the two rings are distinct, Splice will combine them
// into one; (ii) if the two are the same ring, Splice will break it into two
// separate pieces. 
// Thus, Splice can be used both to attach the two edges together and
// to break them apart.
{
	Edge* alpha = a->Onext()->Rot();
	Edge* beta = b->Onext()->Rot();

	Edge* t1 = b->Onext();
	Edge* t2 = a->Onext();
	Edge* t3 = beta->Onext();
	Edge* t4 = alpha->Onext();

	a->next = t1;
	b->next = t2;
	alpha->next = t3;
	beta->next = t4;
}

void DeleteEdge(Edge* e)
{
	Splice(e, e->Oprev());
	Splice(e->Sym(), e->Sym()->Oprev());
	delete e->Qedge();
}

/******************* Topological Operations for Delaunay Diagrams ********************/

Subdivision::Subdivision(const Point2d& a, const Point2d& b, const Point2d& c)
// Initialize a subdivision to the triangle defined by the points a, b, c.
{
	Point2d *da, *db, *dc;
	da = new Point2d(a);
	db = new Point2d(b);
	dc = new Point2d(c);

	Edge* ea = MakeEdge();
	ea->EndPoints(da, db);
	Edge* eb = MakeEdge();
	Splice(ea->Sym(), eb);
	eb->EndPoints(db, dc);
	Edge* ec = MakeEdge();
	Splice(eb->Sym(), ec);
	ec->EndPoints(dc, da);
	Splice(ec->Sym(), ea);
	startingEdge = ea;
}

Edge* Connect(Edge* a, Edge* b)
// Add a new edge e connecting the destination of a to the
// origin of b, in such a way that all three have the same
// left face after the connection is complete.
// Additionally, the data pointers of the new edge are set.
{
	Edge* e = MakeEdge();
	Splice(e, a->Lnext());
	Splice(e->Sym(), b);
	e->EndPoints(a->Dest(), b->Org());
	return e;
}

void Swap(Edge* e)
// Essentially turns edge e counterclockwise inside its enclosing
// quadrilateral. The data pointers are modified accordingly.
{
	Edge* a = e->Oprev();
	Edge* b = e->Sym()->Oprev();
	Splice(e, a);
	Splice(e->Sym(), b);
	Splice(e, a->Lnext());
	Splice(e->Sym(), b->Lnext());
	e->EndPoints(a->Dest(), b->Dest());
}

/******************* Geometric Predicates for Delaunay Diagrams ********************/

inline double TriArea (const Point2d& a, const Point2d& b, const Point2d& c)
// Return twice the area of the oriented triangle (a, b, c), i.e., the
// area is positive if the triangle is oriented counterclockwise.
{
	return (b.x - a.x)*(c.y - a.y) - (b.y - a.y)*(c.x - a.x);
}

int InCircle(const Point2d& a, const Point2d& b,
			 const Point2d& c, const Point2d& d)
// Returns true if the point d is inside the circle defined by the
// points  a, b and c
{
	return (a.x*a.x + a.y*a.y) * TriArea(b, c, d) -
		   (b.x*b.x + b.y*b.y) * TriArea(a, c, d) +
		   (c.x*c.x + c.y*c.y) * TriArea(a, b, d) -
		   (d.x*d.x + d.y*d.y) * TriArea(a, b, c) > 0;
}

int ccw(const Point2d& a, const Point2d& b, const Point2d& c)
// Returns TRUE if the points a, b, c are in a counterclockwise order
{
	return (TriArea(a, b, c) > 0);
}

int RightOf(const Point2d& x, Edge* e)
{
	return ccw(x, e->Dest2d(), e->Org2d());
}

int LeftOf(const Point2d& x, Edge* e)
{
	return ccw(x, e->Org2d(), e->Dest2d());
}

int OnEdge(const Point2d& x, Edge* e)
// A predicate that determines if the point x is on the edge e.
// The point is considered on if it is in the EPS(ilon)-neighbourhood
// of the edge
{
	double ratio1 = (e->Org2d().x - e->Dest2d().x) / (e->Org2d().y - e->Dest2d().y);
	double ratio2 = (x.x - e->Dest2d().x) / (x.y - e->Dest2d().y);

	if (abs(ratio1-ratio2) < EPS)
		return true;
	else
		return false;


	//double t1, t2, t3;
	//t1 = (x - e->Org2d()).norm();
	//t2 = (x - e->Dest2d()).norm();
	//if (t1 < EPS || t2 < EPS)
	//	return true;
	//t3 = (e->Org2d() - e->Dest2d()).norm();
	//if (t1 > t3 || t2 > t3)
	//	return false;
	//Line line(e->Org2d(), e->Dest2d());
	//return fabs(line.eval(x)) < EPS);
}

/******************* An incremental Algorithm fo the construction of ********************/
/**************************** Delaunay Diagrams ****************************************/

Edge* Subdivision::Locate(const Point2d& x)
// returns an edge em such that either x is on e, or e is an edge of a
// triangle containing x. The search starts from startingEdge
// and proceeds in the general direction of x.
{
	Edge* e = startingEdge;

	while (true)
	{
		if (x == e->Org2d() || x == e->Dest2d())
			return e;
		else if (RightOf(x, e))
				e = e->Sym();
		else if (!RightOf(x, e->Onext()))
				e = e->Onext();
		else if (!RightOf(x, e->Dprev()))
				e = e->Dprev();
		else return e;
	}
}

void Subdivision::InsertSite(const Point2d& x)
// Inserts a new point into a subdivision representing a Delaunay
// triangulation, and fixes the affected edges so that the result
// is still a Delanauy triangulation.
{
	Edge* e = Locate(x);
	if ((x == e->Org2d()) || (x == e->Dest2d())) // point is already in
		return;
	else if (OnEdge(x, e))
	{
		e = e->Oprev();
		DeleteEdge(e->Onext());
	}

	// connect the new point to the vertices of the containing
	// triangle (or quadrilateral, if the new point fell on an
	// esisting edge)

	Edge* base = MakeEdge();
	base->EndPoints(e->Org(), new Point2d(x));
	Splice(base, e);
	startingEdge = base;
	do {
		base = Connect(e, base->Sym());
		e  = base->Oprev();
	} while (e->Lnext() != startingEdge);

	// Examine suspect edges to ensure that the delanauy condition is 
	// satisfied

	do {
		Edge* t = e ->Oprev();
		if (RightOf(t->Dest2d(), e) &&
			InCircle(e->Org2d(), t->Dest2d(), e->Dest2d(), x))
		{
			Swap(e);
			e = e->Oprev();
		}
		else if (e->Onext() == startingEdge) // no more suspect edges
			return;
		else		// pop a suspect edge
			e = e->Onext()->Lprev();
	} while (true);
}


int APIENTRY WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
{
  cIFS_Interface iface;

  try
  {
    Console console;
    Format format(32,0x00FF0000,0x0000FF00,0x000000FF);
	console.option("DirectX");
	console.option("default width 800");
	console.option("default height 600");
	//console.option("windowed output");
    console.open("IFS Edit",800, 600, format);
    Surface surface(800, 600, format);

	Screen page;
	
	page.cx = 800;
	page.cy = 600;
	page.clipLeft = 0;
	page.clipRight = 799;
	page.clipUp = 0;
	page.clipDown = 599;

	int frame_nr = 0;

	unsigned char colors[4096 * 4];

//	int32* background = new int32[800*600];

//	Calc_Mandelbrot(background, 800, 600, 0, 0, 800, 600,
//			            -0.755104, -0.7408, 0.10511, 0.11536,
//						colors, 1000, 0,  128);

//	Point2d above(0, -1000),
//			below_left(-1000, 1000),
//			below_right(+1000, 1000);
//
//	Subdivision divide (above, below_left, below_right);
//
//	Point2d p1(10, 20), p2(20, 50), p3(10, 100), p4(100, 100);
//
//
//	divide.InsertSite(p1);
//	divide.InsertSite(p2);
//	divide.InsertSite(p3);
//	divide.InsertSite(p4);
 
	Key toets(PTC_KEY_ENTER);
    while (toets != PTC_KEY_ESCAPE)
    {
	
	  frame_nr ++;

	  int i;
	  for (i = 0; i < 4096; i++)
	  {
		colors[i*4] = abs(((int)(i * 3.5) % 400)-200) + 55;
		colors[i*4+1] = abs(((int)(i * 8.3)% 400)-200) + 55;
		colors[i*4+2] = abs(((int)(i * 2.8)% 400)-200) + 55;	  
	  }
	  while (console.key())
	  { toets = console.read();
//	    iface.Handle_Key(toets);
	  }

	  page.start = (int32*)surface.lock();

	  if (toets != PTC_KEY_PAUSE)
	  { 	    
//		iface.Draw(&page);

		Calc_Mandelbrot(page.start, 800, 600, 0, 0, 800, 600,
			            -2.75, 1.75, -1.5, 1.5,
						colors, 1000, 0,  250);

//		Calc_Mandelbrot(page.start, 800, 600, 0, 0, 800, 600,
//			            -0.755104, -0.7408, 0.10511, 0.11536,
//						colors, 1000, 0,  250);

//		Calc_Mandelbrot(page.start, 800, 600, 0, 0, 800, 600,
//			            -0.74591, -0.74448, 0.11196, 0.11339,
//						colors, 3000, 0,  250);

//		 delete[] background;

	  }

	  surface.unlock();
      surface.copy(console);
      console.update();
    }
  }
  catch (Error &error)
  { error.report(); }

 
  return 0;
}
