///////////////////////////////////////////////////////////////////////
// class cIFS_Transform: a single transformation plus the
//   associated color
///////////////////////////////////////////////////////////////////////

#include "IFS_Transform.h"
#include "affine.h"
#include "math.h"

cIFS_Transform::cIFS_Transform()
{ 
  Reset();
}

void
cIFS_Transform::Reset()
{
  make_normal_mat(mRotate);
  make_normal_mat(mAnim.mRotate);
  for (int i = 0; i < 3; i++)
  { mScale[i] = 0.5;
    mTrans[i] = 0;
	mSkew[i] = 0;	
	mRotate[i*4] = 1;
  }
}

cIFS_Transform::cIFS_Transform(ulong color)
{ cIFS_Transform();
  mColor = color;
}

void
cIFS_Transform::Transform(int inAction, int inAxis, int inAmount)
{
	if (inAction == eScale)
		mScale[inAxis] += 0.01*inAmount;
	else
	if (inAction == eRotate)
	{
		double theMatrix[9];
		if  (inAxis == eX)
			make_rotx_mat(theMatrix, inAmount*MATH_PI/128);
		else
		if  (inAxis == eY)
			make_roty_mat(theMatrix, inAmount*MATH_PI/128);
		else		
			make_rotz_mat(theMatrix, inAmount*MATH_PI/128);
		
        mul_mat_self_mat(theMatrix, mRotate);
		Gram_Schmid_orthonormalize(&(mRotate[0]),&(mRotate[3]), &(mRotate[6]));
	}
	else
	if (inAction == eSkew)
	{
		if (((inAmount < 0) && (mSkew[inAxis] > - MATH_PI/2)) ||
		    ((inAmount > 0) && (mSkew[inAxis] < MATH_PI/2)))
		    mSkew[inAxis] += inAmount*MATH_PI/128;
	}
	else	
		mTrans[inAxis] += 5*inAmount;
}

void
cIFS_Transform::SetScale( double* inScale) 
{
	for (int i = 0; i < 3; i++)
		mScale[i] = inScale[i];
}

void
cIFS_Transform::SetRotate( double* inRotate) 
{
	for (int i = 0; i < 9; i++)
		mRotate[i] = inRotate[i];
}

void
cIFS_Transform::SetSkew( double* inSkew) 
{
	for (int i = 0; i < 3; i++)
		mSkew[i] = inSkew[i];
}

void
cIFS_Transform::SetTrans( double* inTrans) 
{
	for (int i = 0; i < 3; i++)
		mTrans[i] = inTrans[i];
}

double
cIFS_Transform::CalcLimit()
{
	return 1000 * mScale[0]*cos(mSkew[0])*
				  mScale[1]*cos(mSkew[1])*
				  mScale[2]*cos(mSkew[2]);
}

////////////////////////////////////
// animates the current transformation according to its animation settings and
// the frame nr, using sinus-interpolation
////////////////////////////////////
void cIFS_Transform::Animate(ulong frame_nr)
{ 
  double temp_mat[9], amplitude, cycle_length, displacement, value;
  int position, i, j;

// scaling, skewing and translating
  for (j = 0; j < 4; j++)
  { 
    if (j == 1) continue; //skip rotating

	for (i = 0; i < 3; i++) // x, y, z
	{
	  position = frame_nr % mAnim.mCycle[j][i]; // get position in current cycle
	  if (mAnim.mSine[j][i]) // if sine interpolation
	  { 
		amplitude = (mAnim.mLimit[j][i] - mAnim.mBase[j][i]) / 2;
		cycle_length = 2* MATH_PI / mAnim.mCycle[j][i];
		displacement = mAnim.mCycle[j][i] / 4;

		value = amplitude * sin(cycle_length * (position - displacement)) + amplitude;
	  }
	  else
	  { if (position > mAnim.mCycle[j][i]>>1)
	      position = mAnim.mCycle[j][i] - position;
	    value = (position * (mAnim.mLimit[j][i] - mAnim.mBase[j][i]))/(mAnim.mCycle[j][i]>>1);
	  }

	  switch (j){
		case 0: mScale[i] = mAnim.mBase[0][i] + value; break;
		case 1: break;
		case 2: mSkew[i] = mAnim.mBase[2][i] + value; break;
		case 3: mTrans[i] = mAnim.mBase[3][i] + value; break;
	  }
	}
  }

// rotating

  for (i = 0; i < 9; i++)
	mRotate[i] = mAnim.mRotate[i];

  for (i = 0; i < 3; i++) // x, y, z
	if (mAnim.mLimit[1][i] != 0)
	{
	  position = frame_nr % mAnim.mCycle[1][i]; // get position in current cycle
	  amplitude = mAnim.mLimit[1][i] / 2;

	  // circular motion
	  if ((fabs(amplitude - MATH_PI) <= 0.001) ||
		  (fabs(amplitude + MATH_PI) <= 0.001))
		{
		  if (mAnim.mLimit[1][i] < 0) value = - position*2*MATH_PI/mAnim.mCycle[1][i];
								   else value = position*2*MATH_PI/mAnim.mCycle[1][i];
		}
	  else
	  if (mAnim.mSine[1][i]) // sine interpolation
	  {
		cycle_length = 2* MATH_PI / mAnim.mCycle[1][i];
		displacement = mAnim.mCycle[1][i] / 4;
		value = amplitude * sin(cycle_length * (position - displacement)) + amplitude;
	  }
	  else // lineair interpolation
	  { if (position > mAnim.mCycle[1][i]>>1)
		  position = mAnim.mCycle[1][i] - position;
		value = (position * mAnim.mLimit[1][i])/(mAnim.mCycle[1][i]>>1);
	  }

	  if (i == 0) make_rotx_mat(temp_mat, value);
	  if (i == 1) make_roty_mat(temp_mat, value);
	  if (i == 2) make_rotz_mat(temp_mat, value);

	  mul_mat_mat_self(temp_mat, mRotate);
	  Gram_Schmid_orthonormalize(&(mRotate[0]),
			  					 &(mRotate[3]),
			  					 &(mRotate[6]));
	}
}

void cIFS_Transform::SwitchInterpol(int command)
{ mAnim.mSine[command/3][command%3] = !mAnim.mSine[command/3][command%3];
}

void cIFS_Transform::SetBase(int command)
{ int i,
	  com = command / 3,
      dim = command % 3;

  switch (com) {
    case 0: mAnim.mBase[0][dim] = mScale[dim]; break;
	case 1: for (i = 0; i < 9; i++) mAnim.mRotate[i] = mRotate[i]; break;
	case 2: mAnim.mBase[2][dim] = mSkew[dim]; break;
	case 3: mAnim.mBase[3][dim] = mTrans[dim]; break;
  }
}

void cIFS_Transform::SetLimit(int command)
{ int com = command / 3,
      dim = command % 3;

  switch (com) {
    case 0: mAnim.mLimit[0][dim] = mScale[dim]; break;
	case 1: if (mAnim.mLimit[1][dim] == 2*MATH_PI)
				 mAnim.mLimit[1][dim] = 0;
			else mAnim.mLimit[1][dim] = 2*MATH_PI; break;
	case 2: mAnim.mLimit[2][dim] = mSkew[dim]; break;
	case 3: mAnim.mLimit[3][dim] = mTrans[dim]; break;
  }
}

void cIFS_Transform::SetFullBase(void)
{ for (int i = 0; i < 12; i++)
    SetBase(i);
}

void cIFS_Transform::SetFullLimit(void)
{ int i;
  for (i = 0; i < 3; i++) SetLimit(i);
  for (i = 6; i < 12; i++) SetLimit(i);
}

void cIFS_Transform::IncLimit(int command)
{ int com = command / 3,
      dim = command % 3;

  switch (com) {
    case 0: mAnim.mLimit[0][dim] += 0.01; break;
	case 1: if ((mAnim.mLimit[1][dim] - 2*MATH_PI) > -0.001 &&
				(mAnim.mLimit[1][dim] - 2*MATH_PI) <  0.001)
				  mAnim.mLimit[1][dim] = -2*MATH_PI;
			else  mAnim.mLimit[1][dim] += MATH_PI/64; break;
	case 2: mAnim.mLimit[2][dim] += MATH_PI/64; break;
	case 3: mAnim.mLimit[3][dim] += 5; break;
  }
}

void cIFS_Transform::DecLimit(int command)
{ int com = command / 3,
      dim = command % 3;

  switch (com) {
    case 0: mAnim.mLimit[0][dim] -= 0.01; break;
	case 1: if ((mAnim.mLimit[1][dim] + 2*MATH_PI) > -0.001 &&
				(mAnim.mLimit[1][dim] + 2*MATH_PI) <  0.001)
				  mAnim.mLimit[1][dim] = 2*MATH_PI;
			else  mAnim.mLimit[1][dim] -= MATH_PI/64; break;
	case 2: mAnim.mLimit[2][dim] -= MATH_PI/64; break;
	case 3: mAnim.mLimit[3][dim] -= 5; break;
  }
}

void cIFS_Transform::IncCycle(int command)
{
  mAnim.mCycle[command / 3][command % 3]++;
}

void cIFS_Transform::DecCycle(int command)
{
  int com = command / 3,
      dim = command % 3;

  if(mAnim.mCycle[com][dim] > 1) mAnim.mCycle[com][dim]--;
}